import logging
logging.basicConfig(
    level=logging.WARN,
    format='%(levelname)-7.7s %(module)s.py:%(lineno)s \t%(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)
log = logging.getLogger(__name__)
import base64
import click
import os
from datetime import datetime
from pprint import pprint
import pbaccount.api.core as coreapi


@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Enable verbose output. Can be used twice to increase verbosity.",
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
)
@click.option(
    "-c",
    "--config",
    type=click.File("r"),
    default="/etc/psybook.ini",
)
@click.group()
@click.pass_context
def main(ctx, verbose=False, dry_run=False, config=None):
    if verbose:
        if verbose == 1:
            logging.getLogger().setLevel(logging.INFO)
            log.setLevel(logging.INFO)
        elif verbose > 1:
            logging.getLogger().setLevel(logging.DEBUG)
            log.setLevel(logging.DEBUG)
    if not config:
        click.echo("Missing configuration file", err=True)
        raise click.BadOptionUsage("config", "No configuration found")
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    ctx.obj["dry-run"] = dry_run
    ctx.obj["config"] = config.name

@main.group()
@click.pass_context
def invitation(ctx):
    """Manage pending invitations"""
    pass


@invitation.command(name="list")
@click.pass_context
def invitation_list(ctx):
    """List pending invitations"""
    cfg = ctx.obj["config"]
    for i in coreapi.list_invitations(cfg=cfg):
        click.echo("{:%Y-%m-%d} {} {}".format(i.expire, i.token, i.email))


@invitation.command(name="new")
@click.argument("email")
@click.pass_context
def invitation_new(ctx, email):
    """Generate a unique token and send an invitation email"""
    cfg = ctx.obj["config"]
    res = coreapi.create_invitation(email, cfg=cfg, dry_run=ctx.obj.get("dry-run"))
    log.info(res)


@invitation.command(name="cancel")
@click.argument("token")
@click.pass_context
def invitation_cancel(ctx, token, **kwargs):
    """Remove a pending invitation"""
    cfg = ctx.obj["config"]
    coreapi.remove_invitation(token, cfg=cfg, dry_run=ctx.obj.get("dry-run"))


@invitation.command(name="remove-expired")
@click.pass_context
def invitation_remove_expired(ctx):
    """Remove all expired invitations"""
    cfg = ctx.obj["config"]
    coreapi.remove_expired_invitations(cfg=cfg, dry_run=ctx.obj.get("dry-run"))


@main.group()
@click.pass_context
def account(ctx):
    """Handle psybook accounts"""
    pass


@account.command(name="register")
@click.argument("email")
@click.argument("token")
@click.argument("nickname", required=False)
@click.argument("password", required=False)
@click.pass_context
def account_register(ctx, email, token, nickname=None, password=None):
    """Register an account with an invitation token"""
    cfg = ctx.obj["config"]
    invit = coreapi.find_invitation(token, cfg)
    if invit is None:
        raise click.ClickException("The invitation token does not exist.")
    if nickname is None:
        nickname = click.prompt("Choose a nickname")
    if password is None:
        log.warning("Using the token as password. Change this ASAP!")
        password = token
    res = coreapi.create_account(nickname, email, password, cfg=cfg)
    if not res:
        raise click.ClickException("Failed to create account in LDAP. {}".format(res))
    coreapi.remove_invitation(token, cfg)


@account.command(name="remove")
@click.argument("nickname", required=False)
@click.pass_context
def account_remove(ctx, nickname=None):
    cfg = ctx.obj["config"]
    pprint(coreapi.remove_account(nickname, cfg=cfg))

@account.command(name="show")
@click.argument("nickname")
@click.option(
    "-d",
    "--discourse",
    is_flag=True,
)
@click.pass_context
def account_show(ctx, nickname, **kwargs):
    """Print account details"""
    account = coreapi.get_account(nickname, cfg = ctx.obj["config"], verbose = ctx.obj.get("verbose"), **kwargs)
    if account.get("jpegPhoto"):
        account["jpegPhoto"] = base64.b64encode(account.get("jpegPhoto")).decode()[0:15] + "..."
    pprint(account)


@account.command(name="list")
@click.option(
    "-d",
    "--discourse",
    is_flag=True,
)
@click.pass_context
def account_list(ctx, **kwargs):
    """Print a list of existing psybook accounts"""
    for u in coreapi.list_accounts(cfg = ctx.obj["config"], **kwargs):
        print("{uid:25} {mail}".format(**u))


@account.command(name="set")
@click.argument("uid")
@click.argument("itemname")
@click.argument("itemvalue")
@click.pass_context
def account_set(ctx, uid, itemname, itemvalue):
    """
    Update one item of a user's profile.

    \b
    Possible values for itemname:
    - description
    - email
    """
    coreapi.profile_update_item(uid, itemname, itemvalue, cfg = ctx.obj["config"])


@account.command(name="sync-discourse")
@click.argument("uid")
@click.pass_context
def account_sync_discourse(ctx, uid):
    """Sync user's LDAP profile into user's Discourse profile."""
    coreapi.profile_update(uid, cfg = ctx.obj["config"])


@main.group()
@click.pass_context
def discourse(ctx):
    """Interact with Discourse API"""
    pass


@discourse.command(name="global-settings")
@click.pass_context
def discourse_settings(ctx):
    """Print discourse settings"""
    discoursecli = coreapi._open_discourse_cnx(ctx.obj["config"])
    pprint(discoursecli.site_settings())

@discourse.command(name="set")
@click.argument("setting_name")
@click.argument("value")
@click.pass_context
def discourse_settings(ctx, setting_name, value):
    """Change a discourse setting"""
    discoursecli = coreapi._open_discourse_cnx(ctx.obj["config"])
    pprint(discoursecli.update_site_setting(setting_name, value))
