import logging

log = logging.getLogger(__name__)
import base64
from datetime import datetime, timedelta
import json
from pyramid.paster import setup_logging, get_appsettings
from pyramid.threadlocal import get_current_request

from ldap3 import Server, Connection, ALL, SUBTREE, SYNC, HASHED_SALTED_SHA, MODIFY_ADD, MODIFY_REPLACE
from ldap3.utils.hashed import hashed

from secrets import token_urlsafe
from pbaccount.models.allmodels import Invitation
from .discourseapi import DiscourseClient


def create_account(uid, mail, password, cfg=None):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    cnx = _open_ldapcnx(cfg)

    hashed_password = hashed(HASHED_SALTED_SHA, password)
    new_user_dn = "uid={uid},ou=people,{base_dn}".format(uid=uid, base_dn=cfg.get("ldap.base_dn"))
    log.debug("Inserting {} with mail={}".format(new_user_dn, mail))
    ret = cnx.add(
        new_user_dn,
        object_class=["aphmPeople", "top"],
        attributes={
            "mail": mail,
            "userPassword": [hashed_password],
            "sn": uid, # XXX hard-coded value
        }
    )
    cnx.modify(
        "cn=caregiver,ou=groups,o=aphm",
        {
            "uniqueMember": [(MODIFY_ADD, new_user_dn)]
        }
    )
    cnx.unbind()
    # Also create profile in discourse
    discoursecli = _open_discourse_cnx(cfg)
    discoursecli.create_user(uid, uid, mail, hashed_password + "_we_dont_care")
    return ret


def remove_account(uid, cfg=None):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    cnx = _open_ldapcnx(cfg)
    ret = {}
    ret["ldap"] = cnx.delete("uid={},ou=people,o=aphm".format(uid))
    discoursecli = _open_discourse_cnx(cfg)
    try:
        ret["discourse"] = discoursecli.delete_user(uid)
    except Exception as e:
        log.debug(e)
        ret["discourse"] = False
    return ret

def get_account(uid, cfg=None, **kwargs):
    if kwargs.get("discourse", False):
        discoursecli = _open_discourse_cnx(cfg)
        r = discoursecli.get_userinfo(uid)
        if kwargs.get("verbose"):
            return r
        _keys = set(("id", "name", "username", "email", "bio_raw", "custom_fields"))
        res = {k:v for k,v in r.items() if k in _keys}
        return res

    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    cnx = _open_ldapcnx(cfg)
    cnx.search(
        search_base=cfg.get("ldap.base_dn"),
        search_filter="(&(objectClass=aphmPeople)(uid={}))".format(uid),
        search_scope=SUBTREE,
        attributes=[
            "uid",
            "mail",
            "cn",
            "description",
            "mobile",
            "jpegPhoto",
            "aphmPatientDetails",
            "aphmExperience",
        ],
    )
    if len(cnx.entries) != 1:
        return None
    r = cnx.entries[0]
    res = _ldap_entry_as_dict(r.entry_attributes_as_dict)

    # Parse attributes "aphmPatientDetails"
    _aphmPatientDetails = []
    raw_aphmPatientDetails = res.get("aphmPatientDetails", {})
    if raw_aphmPatientDetails:
        if isinstance(raw_aphmPatientDetails, str):
            raw_aphmPatientDetails = [raw_aphmPatientDetails]
        for d in raw_aphmPatientDetails:
            patient = {}
            for subattr in d.strip("[]").split("]["):
                k,v = subattr.split("=")
                patient[k] = v
            _aphmPatientDetails.append(patient)
    res["aphmPatientDetails"] = _aphmPatientDetails
    return res


def get_account_by_mail(mail, cfg=None):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    cnx = _open_ldapcnx(cfg)
    cnx.search(
        search_base=cfg.get("ldap.base_dn"),
        search_filter="(&(objectClass=aphmPeople)(mail={}))".format(mail),
        search_scope=SUBTREE,
        attributes=[
            "uid",
        ],
    )
    if len(cnx.entries) != 1:
        return None
    r = cnx.entries[0]
    res = _ldap_entry_as_dict(r.entry_attributes_as_dict)
    return res


def list_accounts(cfg=None, **kwargs):
    if kwargs.get("discourse", False):
        discoursecli = _open_discourse_cnx(cfg)
        r = discoursecli.list_users()

        if kwargs.get("verbose"):
            return r
        for e in r:
            yield {
                "uid": e.get("username"),
                "mail": "",
            }
        return
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    cnx = _open_ldapcnx(cfg)
    cnx.search(
        search_base=cfg.get("ldap.base_dn"),
        search_filter="(&(objectClass=aphmPeople))",
        search_scope=SUBTREE,
        attributes=[
            "uid",
            "mail",
            "cn",
            "description",
            "mobile",
            "jpegPhoto",
        ],
    )
    for r in cnx.entries:
        res = _ldap_entry_as_dict(r.entry_attributes_as_dict)
        if res.get("jpegPhoto"):
            res["jpegPhoto"] = base64.b64encode(res.get("jpegPhoto")).decode()
        yield res


def create_invitation(email, cfg=None, dry_run=False):
    import smtplib, ssl

    # Make this function callable even out of pyramid's application context
    if cfg is not None:
        settings = _get_settings(cfg)
    else:
        settings = get_current_request().registry.settings
    dbsession = _open_dbsession(cfg)

    # Avoid creating duplicate invitations
    existing = find_invitation_by_email(email, cfg=cfg)
    if existing:
        raise Exception("This email address has already an invitation token: {}".format(existing.token))

    # Avoid inviting someone who already has an account
    existing = get_account_by_mail(email, cfg=cfg)
    if existing:
        raise Exception("This email already has an account: {}".format(existing.get("uid")))

    # Set default validity to one month
    expiration_date = datetime.now() + timedelta(days=31)

    sender = settings.get("smtp.sender", "ne-pas-repondre@psybook.fr")
    recipients = [email]
    invitation_token = token_urlsafe(12)
    message = """Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 8bit
From: {sender}
To: {recipient}
Subject: Invitation Psybook
Vous avez reçu une invitation sur Psybook.

Pour finaliser la création de votre compte et accéder à Psybook, veuillez cliquer sur le lien suivant :

  {site_baseurl}/register/{invitation_token}

-- 
L'équipe Psybook
""".format(
        sender=sender,
        recipient=email,
        invitation_token=invitation_token,
        site_baseurl=settings.get("site.baseurl", "https://www.psybook.fr")
    )
    log.debug("Mail to send:\n{}".format(message))
    if not dry_run:
        model = Invitation(email=email, token=invitation_token, expire=expiration_date)
        dbsession.add(model)
        dbsession.flush()
        if settings.get("smtp.ssl", "false").lower() == "true":
            try:
                context = ssl.create_default_context()
                with smtplib.SMTP_SSL(settings.get("smtp.host"), settings.get("smtp.port"), context=context) as smtpcnx:
                    smtpcnx.login(settings.get("smtp.user"), settings.get("smtp.pass"))
                    smtpcnx.sendmail(sender, recipients, message.encode("utf-8"))
                    log.info("Successfully sent email to {}".format(recipients))
            except Exception as e:
                log.error("Unable to send invitation email", exc_info=(logging.getLogger().level >= logging.DEBUG))
        else:
            try:
                with smtplib.SMTP(settings.get("smtp.host"), settings.get("smtp.port")) as smtpcnx:
                    if settings.get("smtp.user"):
                        smtpcnx.login(settings.get("smtp.user"), settings.get("smtp.pass"))
                    smtpcnx.sendmail(sender, recipients, message.encode("utf-8"))
                    log.info("Successfully sent email to {}".format(recipients))
            except Exception as e:
                log.error("Unable to send invitation email", exc_info=(logging.getLogger().level >= logging.DEBUG))
    return {
        "success": True,
    }


def find_invitation(token, cfg=None):
    dbsession = _open_dbsession(cfg)
    q = dbsession.query(Invitation)\
        .filter(Invitation.token == token)
    return q.first()


def find_invitation_by_email(email, cfg=None):
    dbsession = _open_dbsession(cfg)
    q = dbsession.query(Invitation)\
        .filter(Invitation.email == email)
    return q.first()


def list_invitations(include_expired=False, cfg=None):
    dbsession = _open_dbsession(cfg)
    q = dbsession.query(Invitation)
    if not include_expired:
        q = q.filter(Invitation.expire > datetime.now())
    return q.all()


def remove_invitation(token, cfg=None, dry_run=False):
    dbsession = _open_dbsession(cfg)
    q = dbsession.query(Invitation)\
        .filter(Invitation.token == token)
    if not dry_run:
        q.delete()
        dbsession.flush()


def remove_expired_invitations(cfg=None, dry_run=False):
    dbsession = _open_dbsession(cfg)
    q = dbsession.query(Invitation)\
        .filter(Invitation.expire < datetime.now())
    if not dry_run:
        q.delete()
        dbsession.flush()


def patient_add(uid, pathology, relationship, cfg=None, dry_run=False):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    newvalue = "[pathology={}][relationship={}]".format(pathology, relationship)
    cnx = _open_ldapcnx(cfg)
    cnx.modify(
        "uid={},ou=people,o=aphm".format(uid),
        {
            "aphmPatientDetails": [(MODIFY_ADD, newvalue)]
        }
    )
    cnx.unbind()


def profile_update_item(uid, itemname, itemvalue, cfg=None, dry_run=False):
    """
    Update one profile item into LDAP and into Discourse

    Possible values for itemname:
    - description
    - mail
    - avatar
    - mobile
    - experience (a year)
    - patients (itemvalue is a list of "[pathology=xxx][relationship=yyy]")
    - birthdate
    - department
    - gender
    """
    cfg = _get_settings(cfg)
    discoursecli = _open_discourse_cnx(cfg)
    ldapcli = _open_ldapcnx(cfg)
    if itemname == "description":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "description": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        ldapcli.unbind()
        discoursecli.update_profile_item(uid, "bio_raw", itemvalue)
    elif itemname == "email":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "mail": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        discoursecli.update_email(uid, itemvalue)
    elif itemname == "avatar":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "jpegPhoto": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        discoursecli.update_avatar_image(uid, itemvalue)
    elif itemname == "mobile":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "mobile": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        # We don't want the mobile phone to appear in discourse
    elif itemname == "experience":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "aphmExperience": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        description = _compute_user_description(itemvalue)
        discoursecli.update_profile_item(uid, "custom_fields[description]", description)
    elif itemname == "patients":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "aphmPatientDetails": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        badge_url = "{}//static/img/badge/{}.png".format(
            cfg.get("site.baseurl", "https://www.psybook.fr"),
            _generate_badge_slug(itemvalue),
        )
        discoursecli.update_profile_item(uid, "custom_fields[icon_url]", badge_url)
    elif itemname == "birthdate":
        birthdate = datetime.strptime(itemvalue, '%Y-%m-%d')
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "aphmBirthDate": [(MODIFY_REPLACE, birthdate)]
            }
        )
        # Information not neeeded in discourse
    elif itemname == "gender":
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "aphmGender": [(MODIFY_REPLACE, itemvalue)]
            }
        )
        # Information not neeeded in discourse
    elif itemname == "department":
        department_number = int(itemvalue.split()[0])
        ldapcli.modify(
            "uid={},ou=people,o=aphm".format(uid),
            {
                "aphmDepartment": [(MODIFY_REPLACE, department_number)]
            }
        )
        # Information not neeeded in discourse
    else:
        raise Exception("Unknown item: {}".format(itemname))


def profile_update(uid, cfg=None, dry_run=False):
    """Update one user's profile from LDAP to Discourse"""
    cfg = _get_settings(cfg)
    discoursecli = _open_discourse_cnx(cfg)
    ldap_account = get_account(uid, cfg=cfg)
    badge_url = "{}/static/img/badge/{}.png".format(
        cfg.get("site.baseurl", "https://www.psybook.fr"),
        "-".join(sorted((d.get("pathology") for d in ldap_account.get("aphmPatientDetails")))),
    )

    existing_discourse_account = discoursecli.get_userinfo(uid)
    if not existing_discourse_account:
        fake_password = ldap_account.get("userPassword") + "-" + uid # TODO generate random value instead
        discoursecli.create_account(uid, ldap_account.get("mail"), fake_password)

    discoursecli.update_profile(uid, ldap_account.get("description"), _compute_user_description(ldap_account.get("aphmExperience")), badge_url)
    discoursecli.update_email(uid, ldap_account.get("mail"))
    avatar_data = ldap_account.get("jpegPhoto")
    if avatar_data:
        discoursecli.update_avatar_image(uid, avatar_data)


def _get_settings(config_uri):
    if config_uri is None:
        return get_current_request().registry.settings
    import configparser, os
    cfg = configparser.ConfigParser()
    cfg.read(config_uri)
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, options={"here": os.path.dirname(config_uri)})
    return settings


def _open_dbsession(config_uri=None, section='app:main', **kwargs):
    '''
    Reads SQLAlchemy configuration parameters from the given configuration file
    and creates a new session.
    '''
    if config_uri is None:
        return get_current_request().dbsession
    settings = _get_settings(config_uri)
    from sqlalchemy import engine_from_config
    engine = engine_from_config(settings, "sqlalchemy.")
    from sqlalchemy.orm import sessionmaker
    DBSession = sessionmaker()
    DBSession.configure(bind=engine)
    from ..models.meta import Base
    Base.metadata.bind = engine
    return DBSession()


def _open_ldapcnx(cfg):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    server = Server(cfg.get("ldap.uri"))
    cnx = Connection(
        server,
        cfg.get("ldap.bind_dn"),
        cfg.get("ldap.bind_pw"),
        client_strategy=SYNC,
        auto_bind=True,
    )
    cnx.bind()
    return cnx


def _open_discourse_cnx(cfg):
    if not cfg:
        cfg = get_current_request().registry.settings
    elif isinstance(cfg, str):
        cfg = _get_settings(cfg)
    return DiscourseClient(cfg)


def _ldap_entry_as_dict(entry):
    res = {}
    for k,v in entry.items():
        if isinstance(v, list):
            if len(v) == 1:
                v = v[0]
            elif len(v) == 0:
                v = None
        elif isinstance(v, dict):
            v = _ldap_entry_as_dict(v)
        res[k] = v
    return res


def _deserialize_patient_details(patient_details):
    for d in patient_details:
        patient = {}
        for subattr in d.strip("[]").split("]["):
            k,v = subattr.split("=")
            patient[k] = v
        yield(patient)


def _generate_badge_slug(aphmPatientDetails):
    pathologies = set()
    for d in _deserialize_patient_details(aphmPatientDetails):
        pathologies.add(d.get("pathology"))
    return "-".join(sorted(pathologies))


def _compute_user_description(start_year):
    xp = datetime.now().year - int(start_year)
    return "Aidant(e) depuis {} ans".format(xp)
