#! /usr/bin/env python3

# This python module is largely inspired by pydiscourse.
# cf. https://github.com/bennylope/pydiscourse/blob/v1.1.2/pydiscourse/client.py

import logging

logging.basicConfig(
    level=logging.WARNING, format="%(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)
log = logging.getLogger(__name__)
import requests
from requests.exceptions import HTTPError


class DiscourseError(HTTPError):
    """ A generic error while attempting to communicate with Discourse """


class DiscourseServerError(DiscourseError):
    """ The Discourse Server encountered an error while processing the request """


class DiscourseClientError(DiscourseError):
    """ An invalid request has been made """


class DiscourseRateLimitedError(DiscourseError):
    """ Request required more than the permissible number of retries """


class DiscourseClient(object):
    def __init__(self, cfg, timeout=None):
        self.host = cfg.get("discourse.baseurl")
        self.api_key = cfg.get("discourse.api_key")
        self.api_username = cfg.get("discourse.api_username")
        self.timeout = timeout

    def list_users(self):
        r = self._get("/admin/users/list/active.json", order="created", asc=False)
        return r

    def user_exists(self, username):
        try:
            self.get_userinfo(username)
            return True
        except DiscourseClientError as e:
            return False

    def create_user(self, name, username, email, password, **kwargs):
        """
        Create a Discourse user
        Set keyword argument active='true' to avoid sending activation emails
        TODO: allow optional password and generate a random one
        Args:
            name: the full name of the new user
            username: their username (this is a key... that they can change)
            email: their email, will be used for activation and summary emails
            password: their initial password
            **kwargs: ???? what else can be sent through?
        Returns:
            ????
        """
        if self.user_exists(username):
            return
        r = self._get("/session/hp.json")
        challenge = r["challenge"][::-1]  # reverse challenge, discourse security check
        confirmations = r["value"]
        return self._post(
            "/u",
            name=name,
            username=username,
            email=email,
            password=password,
            password_confirmation=confirmations,
            challenge=challenge,
            **kwargs
        )

    def delete_user(self, username):
        """Delete a user and all the associated posts"""
        try:
            userinfo = self.get_userinfo(username)
            user_id = userinfo.get("user").get("id")
            body_args = {
                "context": "/admin/users/{}/{}".format(user_id, username),
                "delete_posts": "true",
            }
            return self._delete("/admin/users/{}.json".format(user_id), **body_args).get("deleted")
        except DiscourseClientError as e:
            raise

    def get_userinfo(self, username):
        r = self._get("/u/{0}.json".format(username))
        # email address needs an additional request
        r["email"] = self._get("/u/{0}/emails.json".format(username))["email"]
        return r

    def update_email(self, username, email, **kwargs):
        """
        Args:
            username:
            email:
            **kwargs:
        Returns:
        """
        current_email = self._get("/u/{0}/emails.json".format(username))["email"]
        if email == current_email:
            log.debug("Email address is already correct. Ignoring...")
            return None

        # Globally enable email editing in Discourse
        log.debug("Temporarily enabling user email editing in Discourse...")
        self._put("/admin/site_settings/email_editable", email_editable="true")
        # Update user's email address
        log.debug("Setting email for {}: {}".format(username, email))
        ret = self._put(
            "/u/{0}/preferences/email".format(username), email=email, **kwargs
        )
        # Globally disable email editing in Discourse
        log.debug("Disabling user email editing in Discourse...")
        self._put("/admin/site_settings/email_editable", email_editable="false")
        return ret

    def update_profile(self, username, bio, description, aphmBadgeUrl=""):
        """
        Update the profile items relevant for Psybook.
        """
        data = {
            "bio_raw": bio,
            "custom_fields[description]": description,
            "custom_fields[icon_url]": aphmBadgeUrl,
            "external_links_in_new_tab": "true",
        }
        return self._put("/u/{}.json".format(username), **data)

    def update_profile_item(self, username, itemname, itemvalue):
        if itemname in ("mail", "email", "avatar"):
            raise Exception("Unsupported itemname: {}".format(itemname))
        data = {
            itemname: itemvalue,
        }
        return self._put("/u/{}.json".format(username), **data)

    def update_avatar_image(self, username, avatar_data, **kwargs):
        # Step 1 : get userid
        userinfo = self._get("/u/{0}.json".format(username))["user"]
        userid = userinfo.get("id")

        # Step 2 : upload avatar and get upload_id
        files = {
            "files[]": ("avatar_{}.png".format(username), avatar_data, "image/png")
        }
        data = {"user_id": userid, "type": "avatar"}
        resp = self._post("/uploads.json", files=files, **data)
        upload_id = resp.get("id")

        # Step 3 : pick the recently uploaded image as avatar
        data = {"type": "custom", "upload_id": upload_id}
        resp = self._put("/u/{0}/preferences/avatar/pick".format(username), **data)
        return resp

    def trust_level(self, userid, level):
        """
        Update the trust level of a user.
        Adapted from https://github.com/bennylope/pydiscourse/blob/v1.1.2/pydiscourse/client.py#L189-L199

        Args:
            userid: numeric user id
            level: numeric level (0: new_user, ..., 4: leader)
        Returns:
        """
        return self._put("/admin/users/{0}/trust_level".format(userid), level=level)

    def grant_admin(self, userid):
        """
        Turn user into an admin.

        **XXX** this function seems useless. Granting admin privilege
        requires a validation link sent by email.

        Args:
            userid: numeric user id
        Returns:
            {"admin_user":{"id":xx,"username":"..." ....}}
        """
        return self._put("/admin/users/{0}/grant_admin".format(userid))

    def grant_moderation(self, userid):
        """
        Turn user into a moderator.

        Args:
            userid: numeric user id
        Returns:
            {"admin_user":{"id":xx,"username":"..." ....}}
        """
        return self._put("/admin/users/{0}/grant_moderation".format(userid))

    def revoke_moderation(self, userid):
        """
        Turn a moderator into a simple user.

        Args:
            userid: numeric user id
        Returns: nothing
        """
        return self._put("/admin/users/{0}/revoke_moderation".format(userid))

    def update_user(self, username, **kwargs):
        """
        Args:
            username:
            **kwargs:
        Returns:
        """
        return self._put("/users/{0}".format(username), json=True, **kwargs)

    def site_settings(self, *kwargs):
        return self._get("/site.json")


    def update_site_setting(self, setting_name, value):
        kwargs = { setting_name: value }
        return self._put("/admin/site_settings/{}".format(setting_name), **kwargs)


    def _get(self, path, override_request_kwargs=None, **kwargs):
        """
        Args:
            path:
            **kwargs:
        Returns:
        """
        return self._request(
            "GET", path, params=kwargs, override_request_kwargs=override_request_kwargs
        )

    def _put(self, path, json=False, override_request_kwargs=None, **kwargs):
        """
        Args:
            path:
            **kwargs:
        Returns:
        """
        if not json:
            return self._request(
                "PUT",
                path,
                data=kwargs,
                override_request_kwargs=override_request_kwargs,
            )

        else:
            return self._request(
                "PUT",
                path,
                json=kwargs,
                override_request_kwargs=override_request_kwargs,
            )

    def _post(
        self, path, files=None, json=False, override_request_kwargs=None, **kwargs
    ):
        """
        Args:
            path:
            **kwargs:
        Returns:
        """
        if not json:
            return self._request(
                "POST",
                path,
                files=files,
                data=kwargs,
                override_request_kwargs=override_request_kwargs,
            )

        else:
            return self._request(
                "POST",
                path,
                files=files,
                json=kwargs,
                override_request_kwargs=override_request_kwargs,
            )

    def _delete(self, path, override_request_kwargs=None, **kwargs):
        """
        Args:
            path:
            **kwargs:
        Returns:
        """
        return self._request(
            "DELETE",
            path,
            params=kwargs,
            override_request_kwargs=override_request_kwargs,
        )

    def _request(
        self,
        verb,
        path,
        params=None,
        files=None,
        data=None,
        json=None,
        override_request_kwargs=None,
    ):
        """
        Executes HTTP request to API and handles response
        Args:
            verb: HTTP verb as string: GET, DELETE, PUT, POST
            path: the path on the Discourse API
            params: dictionary of parameters to include to the API
            override_request_kwargs: dictionary of requests.request keyword arguments to override defaults
        Returns:
            dictionary of response body data or None
        """
        override_request_kwargs = override_request_kwargs or {}

        url = self.host + path

        headers = {
            "Accept": "application/json; charset=utf-8",
            "Api-Key": self.api_key,
            "Api-Username": self.api_username,
        }

        # How many times should we retry if rate limited
        retry_count = 4
        # Extra time (on top of that required by API) to wait on a retry.
        retry_backoff = 1

        while retry_count > 0:
            request_kwargs = dict(
                allow_redirects=False,
                params=params,
                files=files,
                data=data,
                json=json,
                headers=headers,
                timeout=self.timeout,
            )

            request_kwargs.update(override_request_kwargs)

            response = requests.request(verb, url, **request_kwargs)

            if response.ok:
                break
            if not response.ok:
                #log.debug("request {}".format(response.request.path_url))
                #log.debug("response %s: %s", response.status_code, repr(response.text))
                #log.debug("response.json(): {}".format(response, response.json()))
                try:
                    msg = u",".join(response.json()["errors"])
                except (ValueError, TypeError, KeyError):
                    if response.reason:
                        msg = response.reason
                    else:
                        msg = u"{0}: {1}".format(response.status_code, response.text)

                if 400 <= response.status_code < 500:
                    if 429 == response.status_code:
                        # This codepath relies on wait_seconds from Discourse v2.0.0.beta3 / v1.9.3 or higher.
                        rj = response.json()
                        wait_delay = (
                            retry_backoff + rj["extras"]["wait_seconds"]
                        )  # how long to back off for.

                        if retry_count > 1:
                            time.sleep(wait_delay)
                        retry_count -= 1
                        log.info(
                            "We have been rate limited and waited {0} seconds ({1} retries left)".format(
                                wait_delay, retry_count
                            )
                        )
                        log.debug("API returned {0}".format(rj))
                        continue
                    else:
                        raise DiscourseClientError(msg, response=response)

                # Any other response.ok resulting in False
                raise DiscourseServerError(msg, response=response)

        if retry_count == 0:
            raise DiscourseRateLimitedError(
                "Number of rate limit retries exceeded. Increase retry_backoff or retry_count",
                response=response,
            )

        if response.status_code == 302:
            raise DiscourseError(
                "Unexpected Redirect, invalid api key or host?", response=response
            )

        json_content = "application/json; charset=utf-8"
        content_type = response.headers["content-type"]
        if content_type != json_content:
            # some calls return empty html documents
            if not response.content.strip():
                return None

            raise DiscourseError(
                'Invalid Response, expecting "{0}" got "{1}"'.format(
                    json_content, content_type
                ),
                response=response,
            )

        try:
            decoded = response.json()
        except ValueError:
            raise DiscourseError("failed to decode response", response=response)

        # Checking "errors" length because
        # data-explorer (e.g. POST /admin/plugins/explorer/queries/{}/run)
        # sends an empty errors array
        if "errors" in decoded and len(decoded["errors"]) > 0:
            #log.debug("request {}".format(response.request.path_url))
            #log.debug("response %s: %s", response.status_code, repr(response.text))
            #log.debug("response.json(): {}".format(response, response.json()))

            message = decoded.get("message")
            if not message:
                message = u",".join(decoded["errors"])
            raise DiscourseError(message, response=response)

        return decoded


def read_config(configfile=None):
    """Returns a ``ConfigParser`` object."""
    possible_configfiles = [
        "./discoursecli.ini",
        "~/.config/discoursecli.ini",
        "/etc/discoursecli.ini",
        "/etc/discoursecli/discoursecli.ini",
    ]

    if configfile is None or not os.path.isfile(configfile):
        for candidate in possible_configfiles:
            candidate = os.path.expanduser(candidate)
            if os.path.isfile(candidate):
                configfile = candidate
                log.debug("Found configuration file %s." % configfile)
                break

    if configfile is None:
        raise Exception(
            "No configuration file found. Either use -c option of create one these files: %s"
            % ",".join(possible_configfiles)
        )

    from configparser import ConfigParser

    conf = ConfigParser()
    conf.read(configfile)
    return conf


def parse_args():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print debug messages"
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="veryverbose",
        action="store_true",
        help="Print debug messages",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="Don't really apply modifications, just simulate.",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=read_config,
        default=None,
        help="Read configuration from the specified file.",
    )

    subcommands = parser.add_subparsers(dest="action")

    # subcommand "syncuser"
    cmd_syncuser = subcommands.add_parser("syncuser", help="Update a user profile.")
    cmd_syncuser.add_argument(
        dest="uid", nargs="+", help="List of LDAP uids to update."
    )

    # subcommand "showuser"
    cmd_user = subcommands.add_parser("showuser", help="Managing discourse users info")
    cmd_user.add_argument(dest="uid", nargs="+", help="List of LDAP uids to show.")

    # subcommand "createusers"
    cmd_user = subcommands.add_parser(
        "createusers", help="Create missing users in discourse"
    )
    cmd_user.add_argument(dest="uid", nargs="*", help="List of LDAP uids to create.")

    args = parser.parse_args()
    if args.veryverbose:
        log.setLevel(logging.DEBUG)
        logging.getLogger().handlers[0].setFormatter(
            logging.Formatter(
                "%(levelname)-8.8s %(asctime)-19.19s [%(module)s.py l.%(lineno)s] %(message)s"
            )
        )
    elif args.verbose:
        log.setLevel(logging.INFO)
    return args


if __name__ == "__main__":
    main()
