import logging; log = logging.getLogger(__name__)
from pyramid.session import SignedCookieSessionFactory
from pyramid.authentication import (
    SessionAuthenticationHelper,
)
from pyramid.authorization import (
    ACLHelper,
    Authenticated,
    Everyone,
    Allow,
    ALL_PERMISSIONS,
)

from pbaccount.api.core import get_account


GROUPS = {
    "investigator": [
        "invitation:manager",
        "account:manager",
    ],
}

class PsybookSecurityPolicy:
    def __init__(self):
        self.auth_helper = SessionAuthenticationHelper()
        self.acl = ACLHelper()

    def identity(self, request):
        if request.session.get("userdata"):
            return request.session.get('userdata')
        uid = request.environ.get('REMOTE_USER')
        user = get_account(uid)
        userdata = {
            "userid": uid,
            "principals": ["caregiver"]
        }
        if uid == "eeadmin": # XXX temporary hack, waiting a proper fix of memberOf in slapd
            userdata["principals"] = ["investigator"]
        request.session["userdata"] = userdata
        return userdata

    def authenticated_userid(self, request):
        identity = self.identity(request)
        if identity is not None and "userid" in identity:
            return identity["userid"]

    def remember(self, request, userid, **kw):
        return self.auth_helper.remember(request, userid, **kw)

    def forget(self, request, **kw):
        return self.auth_helper.forget(request, **kw)

    def permits(self, request, context, permission):
        principals = self.effective_principals(request)
        return self.acl.permits(context, principals, permission)

    def effective_principals(self, request):
        principals = [Everyone]
        identity = self.identity(request)
        from pprint import pprint
        if identity is not None and identity.get("userid"):
            principals.append(Authenticated)
            principals.append("u:" + identity["userid"])
            principals.extend(GROUPS.get(identity["userid"], []))
            principals.extend(identity["principals"])
        return principals


def includeme(config):
    settings = config.get_settings()

    my_session_factory = SignedCookieSessionFactory(
        settings.get('session.cookie_signing_secret', 'itsaseekreet'),
        timeout=settings.get('session.timeout', 3600)
    )
    config.set_session_factory(my_session_factory)

    # Trust username from reverse proxy
    config.set_security_policy(PsybookSecurityPolicy())
    config.set_default_permission('authenticated')
    config.add_request_method(get_userdata, 'userdata', reify=True)
    config.set_root_factory(RootFactory)


def get_userdata(request):
    '''Using HTTP session as cache for storing user's information.'''
    if 'userdata' not in request.session:
        user_id = request.unauthenticated_userid
        if user_id is not None:
            user = get_account(user_id)
            request.session['userdata'] = {
                'userid': user.username,
                'principals': [r.get("pathology") for r in user.get("aphmPatientDetails")],
            }
    return request.session.get('userdata')


class RootFactory(object):
    """
    TODO quickly explain how RootFactory is involved in authn/authz mechanisms.
    
    Tip: command line to list all permissions currently expected in view callables and templates.
    
    grep --include=*.py -r 'permission \?=' views/ | sed "s/.*permission \?= \?u\?'\([^']\+\)'.*/\1/" | sort | uniq -c | sort -n
    grep --include=*jinja2 "has_permission" -r templates | sed "s/.*has_permission(u\?'\([^']\+\)').*/\1/" | sort | uniq -c
    """

    def __init__(self, request):
        if not request.path.startswith('/static/'):
            self.__acl__ = [
                (Allow, Authenticated, 'authenticated'),
                (Allow, Authenticated, Authenticated),
                (Allow, 'investigator', ALL_PERMISSIONS),
            ]
        else:
            self.__acl__ = []